/*******************************************************************************

  outer_control.c
  Questions, remarks and complaints may be directed to:  Reid Guenther
                                                         reid@npac.syr.edu

  This file contains an implementation of the outer control variates described
  in the Columbia paper, page 15 (general) and page 18 (specific).

*******************************************************************************/

#include <math.h>
#include <float.h>
#include "path.h"    /* should be included in mesh.h */
#include "option.h"  /* should be included in mesh.h */
#include "mesh.h"    /* mesh data structures */
#include "util.h"

/*
   Note: Q_hat[i] is just the mesh estimated option price from run i,
         u_hat[k][i] is the kth control variate computed given the
                     the meshData for run i,
			u[k] is the actual value that u_hat[k][i] is approximating.

	Note: I assume that the Number_of_Controls is either 1 or 2.  If more
         are desired then some memory allocation will be needed to
         compute all the different average combinations.
*/
double Compute_Outer_Control_Estimate(int Number_of_Controls, 
                                      int Number_of_Runs, double *Q_hat,
                                      double **u_hat, double *u)
{

	int i;
	double Avg_Q_hat, Avg_Q_hat_u1_hat, Avg_u1_hat, Avg_u1_hat_u2_hat;
	double Avg_u2_hat, Avg_Q_hat_u2_hat, Avg_u1_hat_sq, Avg_u2_hat_sq;
	double beta1, beta2, Q_bar_new, a11, a12, a21, a22;
	double a11_inv, a12_inv, a21_inv, a22_inv, rhs1, rhs2;
	double determinant_A;


/*
   	compute the averages:
*/
		Avg_Q_hat = Avg_Q_hat_u1_hat = Avg_u1_hat = 0.0e0;
		for (i=0; i<Number_of_Runs; i++) 
		{
			Avg_Q_hat += Q_hat[i];
			Avg_u1_hat += u_hat[1][i];
			Avg_Q_hat_u1_hat += Q_hat[i]*u_hat[1][i];
			Avg_u1_hat_sq += u_hat[1][i]*u_hat[1][i];
		}

	switch (Number_of_Controls){

		case 1:
			beta1 = (Avg_Q_hat_u1_hat - Avg_Q_hat*Avg_u1_hat)/
						(Avg_u1_hat_sq - Avg_u1_hat*Avg_u1_hat);

			Q_bar_new = Avg_Q_hat - beta1*(Avg_u1_hat - u[1]);

		break;
		case 2:
/* Compute the rest of the averages: */
			Avg_Q_hat_u2_hat = Avg_u1_hat_u2_hat = 0.0e0;
			for (i=0; i<Number_of_Runs; i++)
			{
				Avg_Q_hat_u2_hat += Q_hat[i]*u_hat[2][i];
				Avg_u1_hat_u2_hat += u_hat[1][i]*u_hat[1][i];
			}
/* a11..a22 are elements of the 2x2 matrix that are part of the
  set of linear equations A[][]*beta[] = rhs[] */

			a11 = Avg_u1_hat_sq  - Avg_u1_hat*Avg_u1_hat;
			a12 = a21 = Avg_u1_hat_u2_hat  - Avg_u1_hat*Avg_u2_hat;
			a22 = Avg_u2_hat_sq  - Avg_u2_hat*Avg_u2_hat;
			rhs1 = Avg_Q_hat_u1_hat - Avg_Q_hat*Avg_u1_hat;
			rhs2 = Avg_Q_hat_u2_hat - Avg_Q_hat*Avg_u2_hat;
			determinant_A = a11*a22 - a12*a21;
			a11_inv =  a22/determinant_A;
			a12_inv = -a12/determinant_A;
			a21_inv = -a21/determinant_A;
			a22_inv =  a11/determinant_A;
			beta1 = rhs1*a11_inv + rhs2*a12_inv;
			beta2 = rhs1*a21_inv + rhs2*a22_inv;
			Q_bar_new = Avg_Q_hat - beta1*(Avg_u1_hat - u[1]) -
                              beta2*(Avg_u2_hat - u[2]);

		break;
	}

	return Q_bar_new;
		
}


/*******************************************************************************

   This routine is used to calculate the mesh estimated max european option
   value for the given final time option prices. 

*******************************************************************************/
double OuterCtlEuroMaxMeshValue(meshData *theMesh)
{
	int t, path_i, path_j, Final_t, underlier, Current_t;
	double max_current_price, *tmp;
	double *option_t, *option_tp1, max;

/*
  allocate the t+1 and t maximum european option price arrays:
*/
	Final_t = theMesh->option.pData.times;
	allocate1dDouble(&option_t,   Final_t+1);
	allocate1dDouble(&option_tp1, Final_t+1);

/*
  First find the maximum at each mesh point at the final time
*/
	for (path_i=0; path_i< theMesh->meshPaths;path_i++)
	{
		max =  theMesh->option.pData.price[Final_t][path_i][0];
		for (underlier=1;underlier < theMesh->option.pData.underliers;
		     underlier++)
		{
			if (max < theMesh->option.pData.price[Final_t][path_i][underlier])
				 max = theMesh->option.pData.price[Final_t][path_i][underlier];
		}

		option_tp1[path_i] = max - theMesh->option.strikePrice;
		option_t[path_i]   = 0.0;
	}

		

/*
  Now calculate the mesh estimate of the European option value for the
  previously determined final option values.
*/
	for (Current_t = Final_t - 1; Current_t > 0; Current_t--)
	{
		for (path_i=0; path_i < theMesh->meshPaths; path_i++)
		{
			option_t[path_i] = 0.0;
			for (path_j=0; path_j < theMesh->meshPaths; path_j++)
			{
				option_t[path_i] += theMesh->weights[Current_t][path_i][path_j]*
										  option_tp1[path_j];
			}
		}
/*
	let option_t go to option_tp1
*/
		tmp = option_t;
		option_t = option_tp1;
		option_tp1 = tmp;
	}

	Current_t = 0;
	path_i = 0;
	option_t[path_i] = 0.0;
	for (path_j=0; path_j < theMesh->meshPaths; path_j++)
	{
		option_t[path_i] += theMesh->weights[Current_t][path_i][path_j]*
								  option_tp1[path_j];
	}

	free(option_t);
	free(option_tp1);
	return option_t[path_i]*pow(theMesh->meshPaths,-Final_t);
}
